const std = @import("std");
const zalgebra = @import("zalgebra");
const atlassian = @import("atlassian");

const c = @import("ffi.zig").c;
const gl = @import("gl.zig");
const glutil = @import("glutil.zig");
const GlState = @import("GlState.zig");

pub const FontInfo = struct {
    atlas: atlassian.Atlas(u8),
    freetype: c.FT_Library,
    font_face: c.FT_Face,
    atlas_id: c_uint,
    chars: CharMap,
    line_height: c_long,
    need_reupload: bool,

    pub fn deinit(self: *FontInfo) void {
        self.chars.deinit();
        self.atlas.deinit();
        _ = c.FT_Done_Face(self.font_face);
        _ = c.FT_Done_FreeType(self.freetype);

        self.* = undefined;
    }
};

pub const Character = struct {
    sprite: atlassian.AtlasSprite,
    /// Offset from the left/top
    bearing: zalgebra.Vec2_i32,
    advance: c_long,
};

pub const CharMap = std.AutoHashMap(u21, Character);

// TODO: use fontconfig to find system font?
const font_data = @embedFile("iosevka-regular.ttf");

pub fn load() !FontInfo {
    std.log.info("loading font", .{});
    var freetype: c.FT_Library = undefined;
    if (c.FT_Init_FreeType(&freetype) != 0) {
        return error.FreetypeInit;
    }
    errdefer _ = c.FT_Done_FreeType(freetype);

    var font_face: c.FT_Face = undefined;
    if (c.FT_New_Memory_Face(
        freetype,
        font_data,
        font_data.len,
        0,
        &font_face,
    ) != 0) {
        return error.FontLoading;
    }
    errdefer _ = c.FT_Done_Face(font_face);

    if (c.FT_Set_Pixel_Sizes(font_face, 0, 48) != 0) {
        return error.SetPixelSizes;
    }

    const line_height = (font_face.*.size.*.metrics.ascender -
        font_face.*.size.*.metrics.descender) >> 5;
    var chars = CharMap.init(std.heap.c_allocator);

    var atl = try atlassian.Atlas(u8).init(std.heap.c_allocator, 512 - 64, 512 + 128, 4);
    errdefer atl.deinit();

    var tex: c_uint = 0;
    gl.genTextures(1, &tex);

    gl.bindTexture(gl.TEXTURE_2D, tex);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);

    return .{
        .atlas = atl,
        .freetype = freetype,
        .font_face = font_face,
        .atlas_id = tex,
        .chars = chars,
        .line_height = line_height,
        .need_reupload = false,
    };
}

pub fn getChar(info: *FontInfo, codepoint: u21) !Character {
    if (info.chars.get(codepoint)) |ch| return ch;

    info.need_reupload = true;

    if (c.FT_Load_Char(info.font_face, codepoint, c.FT_LOAD_RENDER) != 0) {
        std.log.warn("freetype failed to load glyph {}", .{codepoint});
        if (codepoint == '?') {
            return error.LoadChar;
        }
        return try getChar(info, '?');
    }

    const bmp = &info.font_face.*.glyph.*.bitmap;

    const sprite = if (bmp.rows * bmp.width == 0)
        atlassian.AtlasSprite{
            .xmin = 0,
            .ymin = 0,
            .xmax = 0,
            .ymax = 0,
        }
    else
        try info.atlas.addSprite(
            bmp.buffer[0 .. bmp.rows * bmp.width],
            bmp.width,
        );

    const char = Character{
        .sprite = sprite,
        .bearing = zalgebra.Vec2_i32.new(
            info.font_face.*.glyph.*.bitmap_left,
            info.font_face.*.glyph.*.bitmap_top,
        ),
        .advance = info.font_face.*.glyph.*.advance.x,
    };

    try info.chars.put(codepoint, char);

    return char;
}

pub fn initFontGl() !void {
    std.log.info("creating font rendering pipeline", .{});

    const vert = try glutil.loadShader(@embedFile("font_vert.glsl"), gl.VERTEX_SHADER);
    defer gl.deleteShader(vert);

    const frag = try glutil.loadShader(@embedFile("font_frag.glsl"), gl.FRAGMENT_SHADER);
    defer gl.deleteShader(frag);

    const program = gl.createProgram();
    gl.attachShader(program, vert);
    gl.attachShader(program, frag);
    gl.linkProgram(program);
    try glutil.checkProgramLinkError(program);

    GlState.state.font_shader_program = program;

    gl.genVertexArrays(1, &GlState.state.font_vao);
    gl.genBuffers(1, &GlState.state.font_vbo);

    gl.bindBuffer(gl.ARRAY_BUFFER, GlState.state.font_vbo);

    gl.bindVertexArray(GlState.state.font_vao);
    gl.vertexAttribPointer(0, 4, gl.FLOAT, gl.FALSE, 4 * @sizeOf(f32), null);
    gl.enableVertexAttribArray(0);
}

pub fn textVertices(
    s: []const u8,
    info: *FontInfo,
    vertex_buf: *std.ArrayList(f32),
    x: f32,
    y: f32,
    scale: f32,
    line_max: f32,
) !void {
    var cur_x = x;
    var cur_y: f32 = 0.0;

    var char_iter = (try std.unicode.Utf8View.init(s)).iterator();
    while (char_iter.nextCodepoint()) |uchar| {
        const ch = try getChar(info, uchar);

        const xpos = cur_x + @as(f32, @floatFromInt(ch.bearing.x())) * scale;
        const ypos = y - (@as(f32, @floatFromInt(
            (@as(i32, @intCast(ch.sprite.ymax - ch.sprite.ymin)) - ch.bearing.y()),
        )) + cur_y) * scale;

        const w = @as(f32, @floatFromInt(ch.sprite.xmax - ch.sprite.xmin)) * scale;
        const h = @as(f32, @floatFromInt(ch.sprite.ymax - ch.sprite.ymin)) * scale;

        var texpos = zalgebra.GenericVector(4, f32).new(
            @as(f32, @floatFromInt(ch.sprite.xmin)) / @as(f32, @floatFromInt(info.atlas.stride)),
            @as(f32, @floatFromInt(ch.sprite.ymin)) / @as(f32, @floatFromInt(@divExact(info.atlas.pixels.len, info.atlas.stride))),
            @as(f32, @floatFromInt(ch.sprite.xmax)) / @as(f32, @floatFromInt(info.atlas.stride)),
            @as(f32, @floatFromInt(ch.sprite.ymax)) / @as(f32, @floatFromInt(@divExact(info.atlas.pixels.len, info.atlas.stride))),
        );

        // zig fmt: off
        try vertex_buf.appendSlice(&[_]f32{
            xpos, ypos + h, texpos.x(), texpos.y(),
            xpos, ypos, texpos.x(), texpos.w(),
            xpos + w, ypos, texpos.z(), texpos.w(),

            xpos, ypos + h, texpos.x(), texpos.y(),
            xpos + w, ypos, texpos.z(), texpos.w(),
            xpos + w, ypos + h, texpos.z(), texpos.y(),
        });
        // zig fmt: on

        // bit shifting stuff is to convert to pixels
        cur_x += @as(f32, @floatFromInt(ch.advance >> 6)) * scale;
        if (cur_x + @as(f32, @floatFromInt(ch.sprite.xmax - ch.sprite.xmin)) >= line_max) {
            cur_x = x;
            cur_y += @as(f32, @floatFromInt(info.line_height)) * scale;
        }
    }
}

pub fn renderText(vertices: []const f32, info: *FontInfo) void {
    gl.bindTexture(gl.TEXTURE_2D, info.atlas_id);

    if (info.need_reupload) {
        std.log.info("reuploading font atlas", .{});
        info.need_reupload = false;

        gl.pixelStorei(gl.UNPACK_ALIGNMENT, 1);

        gl.texImage2D(
            gl.TEXTURE_2D,
            0,
            gl.RED,
            @intCast(info.atlas.stride),
            @intCast(@divExact(info.atlas.pixels.len, info.atlas.stride)),
            0,
            gl.RED,
            gl.UNSIGNED_BYTE,
            info.atlas.pixels.ptr,
        );
    }

    gl.useProgram(GlState.state.font_shader_program);
    gl.bindVertexArray(GlState.state.font_vao);
    gl.bindBuffer(gl.ARRAY_BUFFER, GlState.state.font_vbo);
    gl.bufferData(
        gl.ARRAY_BUFFER,
        @intCast(vertices.len * @sizeOf(f32)),
        vertices.ptr,
        gl.DYNAMIC_DRAW,
    );
    gl.drawArrays(gl.TRIANGLES, 0, @intCast(vertices.len));
    gl.bindTexture(gl.TEXTURE_2D, 0);
}

pub fn renderFontAtlas(info: *const FontInfo) void {
    // zig fmt: off
    const test_vertices = [_]f32{
        0.0, 0.0, 0.0, 1.0,
        0.0, @floatFromInt(@divExact(info.atlas.pixels.len, info.atlas.stride)), 0.0, 0.0,
        @floatFromInt(info.atlas.stride), 0.0, 1.0, 1.0,

        0.0, @floatFromInt(@divExact(info.atlas.pixels.len, info.atlas.stride)), 0.0, 0.0,
        @floatFromInt(info.atlas.stride), 0.0, 1.0, 1.0,
        @floatFromInt(info.atlas.stride), @floatFromInt(@divExact(info.atlas.pixels.len, info.atlas.stride)), 1.0, 0.0,
    };
    // zig fmt: on

    gl.useProgram(GlState.state.font_shader_program);
    gl.bindVertexArray(GlState.state.font_vao);
    gl.bindTexture(gl.TEXTURE_2D, info.atlas_id);
    gl.bindBuffer(gl.ARRAY_BUFFER, GlState.state.font_vbo);
    gl.bufferData(
        gl.ARRAY_BUFFER,
        @intCast(test_vertices.len * @sizeOf(f32)),
        &test_vertices,
        gl.DYNAMIC_DRAW,
    );
    gl.drawArrays(gl.TRIANGLES, 0, @intCast(test_vertices.len));
}
