const std = @import("std");
const zalgebra = @import("zalgebra");
const boxflow = @import("boxflow");
const c = @import("ffi.zig").c;
const gl = @import("gl.zig");
const glutil = @import("glutil.zig");
const Color = @import("Color.zig");
const GlState = @import("GlState.zig");
const font = @import("font.zig");

const BingoBox = struct {
    checked: bool = false,
    b: boxflow.boxes.Simple = .{},
    text: []const u8,

    pub fn render(self: *BingoBox, state: *State) !void {
        if (self.b.data.overflow)
            return;

        if (self.checked) {
            try glutil.rectangle(
                &state.vertex_buf,
                @floatFromInt(self.b.data.pos.x),
                @floatFromInt(
                    GlState.state.current_screen_size.y() -
                        @as(i32, @intCast(self.b.data.pos.y)),
                ),
                @floatFromInt(self.b.data.pos.x + self.b.data.size.width),
                @floatFromInt(
                    GlState.state.current_screen_size.y() -
                        @as(i32, @intCast(self.b.data.pos.y + self.b.data.size.height)),
                ),
                0.0,
                Color{
                    .r = Color.square_frame.r,
                    .g = Color.square_frame.g,
                    .b = Color.square_frame.b,
                    .a = @sin(
                        @as(f32, @floatCast(c.glfwGetTime())),
                    ) / 2.0 + 0.5,
                },
            );
        }

        try glutil.rectangle(
            &state.vertex_buf,
            @floatFromInt(self.b.data.pos.x + 6),
            @floatFromInt(
                GlState.state.current_screen_size.y() -
                    @as(i32, @intCast(self.b.data.pos.y)) - 6,
            ),
            @floatFromInt((self.b.data.pos.x + self.b.data.size.width) -| 6),
            @floatFromInt(
                GlState.state.current_screen_size.y() -
                    @as(i32, @intCast((self.b.data.pos.y + self.b.data.size.height) -| 6)),
            ),
            0.0,
            Color.square_background,
        );

        try font.textVertices(
            self.text,
            &state.font_info,
            &state.font_vertex_buf,
            @floatFromInt(self.b.data.pos.x + 6),
            @as(f32, @floatFromInt(
                GlState.state.current_screen_size.y() -
                    @as(i32, @intCast(self.b.data.pos.y)) - 6,
            )) - 48.0 * 0.4,
            0.4,
            @floatFromInt(self.b.data.pos.x + self.b.data.size.width),
        );
    }
};

const ClearButton = struct {
    b: boxflow.boxes.Simple = .{},

    fn box(self: *ClearButton, arena: std.mem.Allocator) !boxflow.Box {
        const b = try arena.create(boxflow.boxes.Clamped);
        b.* = .{
            .child = self.b.box(),
            .constraints = boxflow.Constraints.tight(.{ .width = 100, .height = 32 }),
        };
        return b.box();
    }

    fn render(self: *ClearButton, state: *State) !void {
        if (self.b.data.overflow)
            return;

        try glutil.rectangle(
            &state.vertex_buf,
            @floatFromInt(self.b.data.pos.x),
            @floatFromInt(GlState.state.current_screen_size.y() -
                @as(i32, @intCast(self.b.data.pos.y))),
            @floatFromInt(self.b.data.pos.x + self.b.data.size.width),
            @floatFromInt(GlState.state.current_screen_size.y() -
                @as(i32, @intCast(self.b.data.pos.y + self.b.data.size.height))),
            0.0,
            Color.clear_button,
        );

        try font.textVertices(
            "Clear",
            &state.font_info,
            &state.font_vertex_buf,
            @floatFromInt(self.b.data.pos.x + 6),
            @as(f32, @floatFromInt(
                GlState.state.current_screen_size.y() -
                    @as(i32, @intCast(state.bottom_bar_box.pos.y)),
            )) / 2.0 - 48.0 * 0.4 / 2.0,
            0.4,
            @floatFromInt(self.b.data.pos.x + self.b.data.size.width),
        );
    }
};

// TODO: make these command line params
const width = 4;
const height = 4;

pub const State = struct {
    vertex_buf: std.ArrayList(f32),
    font_vertex_buf: std.ArrayList(f32),
    boxes: []BingoBox,
    font_info: font.FontInfo,
    prev_mouse_state: c_int = 0,
    debug_render: bool = false,
    bottom_bar_box: boxflow.BoxData = .{},
    file_name: []const u8,
    clear_button: ClearButton = .{},

    pub fn init(squares: [][]const u8, file_name: []const u8) !State {
        const boxes = try std.heap.c_allocator.alloc(BingoBox, width * height);
        errdefer std.heap.c_allocator.free(boxes);

        for (boxes, 0..) |*b, i| {
            b.* = BingoBox{
                .text = squares[i],
            };
        }

        return .{
            .vertex_buf = std.ArrayList(f32).init(std.heap.c_allocator),
            .font_vertex_buf = std.ArrayList(f32).init(std.heap.c_allocator),
            .boxes = boxes,
            .font_info = try font.load(),
            .file_name = file_name,
        };
    }

    pub fn deinit(self: *State) void {
        self.vertex_buf.deinit();
        std.heap.c_allocator.free(self.boxes);
        self.font_info.deinit();

        self.* = undefined;
    }
};

pub fn onKey(state: *State, win: *c.GLFWwindow, key: c_int, action: c_int) void {
    if (action != c.GLFW_PRESS)
        return;

    if (key == c.GLFW_KEY_ESCAPE or key == c.GLFW_KEY_Q) {
        c.glfwSetWindowShouldClose(win, 1);
        return;
    }

    if (key == c.GLFW_KEY_F3) {
        state.debug_render = !state.debug_render;
    }
}

pub fn handleInput(win: *c.GLFWwindow, state: *State) !void {
    // TODO: convert to event-based system
    const mouse_state = c.glfwGetMouseButton(win, c.GLFW_MOUSE_BUTTON_LEFT);
    defer if (state.prev_mouse_state != mouse_state) {
        state.prev_mouse_state = mouse_state;
    };

    if (state.prev_mouse_state != c.GLFW_PRESS and mouse_state == c.GLFW_PRESS) {
        var fmousex: f64 = 0;
        var fmousey: f64 = 0;

        c.glfwGetCursorPos(win, &fmousex, &fmousey);

        const mousex: usize = @intFromFloat(fmousex);
        const mousey: usize = @intFromFloat(fmousey);

        for (state.boxes) |*b| {
            if (mousey > b.b.data.pos.y and mousey < b.b.data.pos.y + b.b.data.size.height and
                mousex > b.b.data.pos.x and mousex < b.b.data.pos.x + b.b.data.size.width)
            {
                b.checked = !b.checked;
                return;
            }
        }

        if (mousey > state.clear_button.b.data.pos.y and
            mousey < state.clear_button.b.data.pos.y + state.clear_button.b.data.size.height and
            mousex > state.clear_button.b.data.pos.x and
            mousex < state.clear_button.b.data.pos.x + state.clear_button.b.data.size.width)
        {
            for (state.boxes) |*b| {
                b.checked = false;
            }
            return;
        }
    }
}

pub fn layout(state: *State, w: c_int, h: c_int) !void {
    var arena_buf: [1024 * 4]u8 = undefined;
    var fba = std.heap.FixedBufferAllocator.init(&arena_buf);
    const arena = fba.allocator();

    var ctx = boxflow.LayoutCtx{ .alloc = std.heap.c_allocator };

    const gridboxes = try arena.alloc(boxflow.Box, state.boxes.len);

    for (state.boxes, 0..) |*b, i| {
        var padding = try arena.create(boxflow.boxes.Padding);
        padding.* = boxflow.boxes.Padding{ .child = b.b.box(), .padding = 4 };

        gridboxes[i] = padding.box();
    }

    var grid = boxflow.boxes.FixedGrid{ .cols = width, .children = gridboxes };

    // needed for FixedGrid to flex properly
    var contracted_grid = boxflow.boxes.Contracted{ .child = grid.box(), .axis = .vertical };
    contracted_grid.data.flex_expand = 1;

    var bar_spacer = boxflow.boxes.Simple{};
    var spacer_cont = boxflow.boxes.Contracted{ .child = bar_spacer.box(), .axis = .horizontal };
    spacer_cont.data.flex_expand = 1;

    var bar_box = try boxflow.boxes.FlowBox.init(
        &ctx,
        .horizontal,
        &.{ spacer_cont.box(), try state.clear_button.box(arena) },
    );
    defer bar_box.deinit();

    var bar_clamp = boxflow.boxes.Clamped{
        .child = bar_box.box(),
        .constraints = .{
            .min = .{ .width = 0, .height = 32 },
            .max = .{ .width = std.math.maxInt(usize), .height = 32 },
        },
    };

    var vbox = try boxflow.boxes.FlowBox.init(
        &ctx,
        .vertical,
        &.{ contracted_grid.box(), bar_clamp.box() },
    );
    defer vbox.deinit();

    try boxflow.layout(
        vbox.box(),
        &ctx,
        .{
            .min = .{ .width = 0, .height = 0 },
            .max = .{ .width = @intCast(w), .height = @intCast(h) },
        },
    );

    state.bottom_bar_box = bar_spacer.data;
}

pub fn draw(win: *c.GLFWwindow, state: *State) !void {
    _ = win;
    state.vertex_buf.clearRetainingCapacity();
    state.font_vertex_buf.clearRetainingCapacity();

    gl.clearColor(
        Color.background.r,
        Color.background.g,
        Color.background.b,
        Color.background.a,
    );
    gl.clear(gl.COLOR_BUFFER_BIT);

    for (state.boxes) |*b| {
        try b.render(state);
    }

    try glutil.rectangle(
        &state.vertex_buf,
        @floatFromInt(state.bottom_bar_box.pos.x),
        @floatFromInt(
            GlState.state.current_screen_size.y() -
                @as(i32, @intCast(state.bottom_bar_box.pos.y)),
        ),
        @floatFromInt((state.bottom_bar_box.pos.x + state.bottom_bar_box.size.width)),
        @floatFromInt(
            GlState.state.current_screen_size.y() -
                @as(i32, @intCast((state.bottom_bar_box.pos.y + state.bottom_bar_box.size.height))),
        ),
        0.0,
        Color.bar_bg,
    );

    var checked_n: usize = 0;
    for (state.boxes) |b| {
        if (b.checked)
            checked_n += 1;
    }

    var textbuf: [1024]u8 = undefined;
    const bar_s = try std.fmt.bufPrint(
        &textbuf,
        "BINGUS - {s} - {} boxes checked",
        .{ state.file_name, checked_n },
    );

    try font.textVertices(
        bar_s,
        &state.font_info,
        &state.font_vertex_buf,
        @floatFromInt(state.bottom_bar_box.pos.x + 6),
        @as(f32, @floatFromInt(
            GlState.state.current_screen_size.y() -
                @as(i32, @intCast(state.bottom_bar_box.pos.y)),
        )) / 2.0 - 48.0 * 0.4 / 2.0,
        0.4,
        std.math.inf(f32),
    );

    try state.clear_button.render(state);

    glutil.drawTriangles(state.vertex_buf.items);
    font.renderText(state.font_vertex_buf.items, &state.font_info);

    // debug stuff
    if (state.debug_render) {
        font.renderFontAtlas(&state.font_info);
    }
}
