const std = @import("std");
const zalgebra = @import("zalgebra");
const c = @import("ffi.zig").c;
const gl = @import("gl.zig");
const glutil = @import("glutil.zig");
const gui = @import("gui.zig");
const GlState = @import("GlState.zig");
const bingoloader = @import("bingoloader.zig");
const font = @import("font.zig");

pub const std_options = struct {
    pub const log_level = .debug;
};

comptime {
    if (@import("builtin").is_test) {
        std.testing.refAllDecls(@This());
    }
}

pub fn main() !void {
    if (std.os.argv.len != 2) {
        std.log.err(
            \\Invalid CLI usage
            \\
            \\Usage:
            \\{s} [bingolist]
        , .{std.os.argv[0]});
        return error.InvalidCLI;
    }

    var bingos = try bingoloader.loadFile(std.mem.span(std.os.argv[1]), 16);
    defer bingos.deinit();

    std.log.info("initializing GL state", .{});
    try GlState.init();
    defer GlState.deinit();

    std.log.info("initializing GLFW", .{});
    if (c.glfwInit() == c.GLFW_FALSE) {
        return error.GlfwInit;
    }
    defer c.glfwTerminate();

    c.glfwWindowHint(c.GLFW_CONTEXT_VERSION_MAJOR, 3);
    c.glfwWindowHint(c.GLFW_CONTEXT_VERSION_MINOR, 3);
    c.glfwWindowHint(c.GLFW_OPENGL_PROFILE, c.GLFW_OPENGL_CORE_PROFILE);

    // fancy transparent window
    c.glfwWindowHint(c.GLFW_TRANSPARENT_FRAMEBUFFER, c.GLFW_TRUE);

    std.log.info("creating window", .{});
    const win = c.glfwCreateWindow(
        GlState.inital_win_size.x(),
        GlState.inital_win_size.y(),
        "Bingus",
        null,
        null,
    );
    if (win == null) {
        return error.GlfwCreateWindow;
    }

    // create OpenGL context
    c.glfwMakeContextCurrent(win);

    std.log.info("loading OpenGL", .{});
    try gl.load({}, getProcAddress);

    // VSync
    c.glfwSwapInterval(1);

    // window resize callback
    _ = c.glfwSetFramebufferSizeCallback(win, &fbResizeCb);

    // mouse button buffering
    c.glfwSetInputMode(win, c.GLFW_STICKY_MOUSE_BUTTONS, c.GLFW_TRUE);

    // key press handler
    _ = c.glfwSetKeyCallback(win, keyEventCb);

    gl.viewport(0, 0, GlState.inital_win_size.x(), GlState.inital_win_size.y());

    std.log.info("loading shaders", .{});
    const vert_shader = try glutil.loadShader(
        @embedFile("vert.glsl"),
        gl.VERTEX_SHADER,
    );
    const frag_shader = try glutil.loadShader(
        @embedFile("frag.glsl"),
        gl.FRAGMENT_SHADER,
    );

    const shader_program = gl.createProgram();
    gl.attachShader(shader_program, vert_shader);
    gl.attachShader(shader_program, frag_shader);
    gl.linkProgram(shader_program);
    try glutil.checkProgramLinkError(shader_program);
    GlState.state.shader_program = shader_program;

    // shaders can be delted once linked to a program
    gl.deleteShader(vert_shader);
    gl.deleteShader(frag_shader);

    gl.genVertexArrays(1, &GlState.state.vao);
    gl.genBuffers(1, &GlState.state.vbo);

    gl.bindVertexArray(GlState.state.vao);
    gl.bindBuffer(gl.ARRAY_BUFFER, GlState.state.vbo);

    // vertex pos
    gl.vertexAttribPointer(0, 3, gl.FLOAT, gl.FALSE, 7 * @sizeOf(f32), null);
    gl.enableVertexAttribArray(0);

    // vertex color
    gl.vertexAttribPointer(
        1,
        4,
        gl.FLOAT,
        gl.FALSE,
        7 * @sizeOf(f32),
        @ptrFromInt(3 * @sizeOf(f32)),
    );
    gl.enableVertexAttribArray(1);

    try font.initFontGl();

    GlState.updateProjectionMatrix(GlState.inital_win_size);

    //gl.enable(gl.CULL_FACE);
    gl.enable(gl.BLEND);
    gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

    //gl.polygonMode(gl.FRONT_AND_BACK, gl.LINE);

    var gui_state = try gui.State.init(
        bingos.bingos,
        std.fs.path.basename(std.mem.span(std.os.argv[1])),
    );
    defer gui_state.deinit();
    try gui.layout(
        &gui_state,
        GlState.inital_win_size.x(),
        GlState.inital_win_size.y(),
    );
    c.glfwSetWindowUserPointer(win, &gui_state);
    while (c.glfwWindowShouldClose(win) == 0) {
        c.glfwPollEvents();
        defer c.glfwSwapBuffers(win);

        try gui.handleInput(win.?, &gui_state);

        if (c.glfwGetWindowAttrib(win, c.GLFW_VISIBLE) == 0) {
            // window isn't visible, so no need to do any rendering
            continue;
        }

        try gui.draw(win.?, &gui_state);
    }
}

fn getProcAddress(_: void, name: [:0]const u8) ?gl.FunctionPointer {
    return c.glfwGetProcAddress(name.ptr);
}

fn keyEventCb(
    win: ?*c.GLFWwindow,
    key: c_int,
    scancode: c_int,
    action: c_int,
    mods: c_int,
) callconv(.C) void {
    _ = scancode;
    _ = mods;
    gui.onKey(
        @ptrCast(@alignCast(c.glfwGetWindowUserPointer(win))),
        win.?,
        key,
        action,
    );
}

fn fbResizeCb(win: ?*c.GLFWwindow, width: c_int, height: c_int) callconv(.C) void {
    const vec = zalgebra.Vec2_i32.new(width, height);
    gl.viewport(0, 0, width, height);
    GlState.updateProjectionMatrix(vec);
    GlState.state.current_screen_size = vec;

    const state: *gui.State = @ptrCast(@alignCast(c.glfwGetWindowUserPointer(win)));

    gui.layout(state, width, height) catch |e| {
        std.log.err("layout: {}", .{e});
    };
}
